from datetime import date
from typing import List

from .base import BasePattern
from .pattern_list import PatternTypes


class Date(BasePattern):
    """This expression will look for a date under the following formats in the text:
    * dd/mm/yyyy
    * dd-mm-yyyy
    * dd.mm.yyyy
    """

    TYPE = PatternTypes.DATE
    REGEX_FORMAT_STRING = (
        r"(?:(?:31(\/|-|\.)(?:0?[13578]|1[02]))\1|(?:(?:29|30)"
        r"(\/|-|\.)(?:0?[13-9]|1[0-2])\2))(?:(?:1[6-9]|[2-9]\d)?\d{2})$|^(?:29(\/|-|\.)"
        r"0?2\3(?:(?:(?:1[6-9]|[2-9]\d)?(?:0[48]|[2468][048]|[13579][26])|(?:(?:16|"
        r"[2468][048]|[3579][26])00))))$|^(?:0?[1-9]|1\d|2[0-8])(\/|-|\.)(?:(?:0?[1-9])"
        r"|(?:1[0-2]))\4(?:(?:1[6-9]|[2-9]\d)?\d{2})"
    )

    def to_date(self, target_string: str) -> List[date]:
        """Converts the parsed dates in a date object

        :param target_string: The string to aprse dates on
        :type target_string: str
        :return: A list of date objects
        :rtype: List[date]
        """

        matches = self.search(target_string)
        dates = []

        for match in matches:
            for sep in "/-.":
                if len(match.split(sep)) == 3:
                    day, month, year = [int(el) for el in match.split(sep)]
                    break

            dates.append(date(year, month, day))

        return dates



import re

from .base import BasePattern
from .pattern_list import PatternTypes


class HyphenWordBreak(BasePattern):
    TYPE = PatternTypes.HYPHEN_WORD_BREAK
    REGEX_FORMAT_STRING = r"[a-zA-Z][\-]$[\n][a-zA-Z]"
    REGEX_SUBSTITUTION_PAIR = (r"([a-zA-Z])[\-]$[\n]([a-zA-Z])", "\\1\\2")

    def reassemble_text(self, target_string: str) -> str:
        regex, subst = self.REGEX_SUBSTITUTION_PAIR
        return re.sub(regex, subst, target_string, 0, re.MULTILINE)



from .base import BasePattern
from .pattern_list import PatternTypes


class CLanguageIdentifier(BasePattern):
    """Looks for identifiers deriving from C within text data.

    Identifiers are also called tokens or symbols, naming language entities, like
    variables, structures, functions, etc...

    It can also be used for languages deriving from C, like C++, Python, Java,
    JavaScript, etc...
    """

    TYPE = PatternTypes.C_IDENTIFIER
    REGEX_FORMAT_STRING = r"\b[a-zA-Z_][0-9a-zA-Z_]*$\b"



from .base import BasePattern
from .pattern_list import PatternTypes


class BitcoinAddress(BasePattern):
    """Looks for any bitcoin address in the text data.

    A bitcoin address must be unique. It's comprised of 25 to 39
    alphanumeric characters and the address usually begins with
    number 1 or 3, or bc1 for bech32 types.

    * P2PKH (Pay-to-Pub key hash) begins with number 1
    * P2SH (Pay-to-Scrypt Hash) begins with number 3
    * Bech32 addresses are a SegWit format starting with bc1
    """

    REGEX_FORMAT_STRING = r"\b(bc1|[13])[a-zA-HJ-NP-Z0-9]{25,39}\b"
    TYPE = PatternTypes.BITCOIN_ADDRESS



from .base import BasePattern
from .pattern_list import PatternTypes


class Url(BasePattern):
    TYPE = PatternTypes.URL
    REGEX_FORMAT_STRING = (
        r"https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\."
        r"[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()!@:%_\+.~#?&\/\/=]*)"
    )



from .base import BasePattern
from .pattern_list import PatternTypes


class Emoji(BasePattern):
    """
    Identifies emojis in the text.

    Emojis can be translated as unicode characters as denoted in the
    unicode standard: https://www.unicode.org/Public/emoji/12.1/emoji-data.txt

    Warning: the emoji pattern wasn't tested properly as some of them had an unexpected
    result
    """

    TYPE = PatternTypes.EMOJI
    REGEX_FORMAT_STRING = (
        r"(\u00a9|\u00ae|[\u2000-\u3300]|\ud83c[\ud000-\udfff]|"
        r"\ud83d[\ud000-\udfff]|\ud83e[\ud000-\udfff])"
    )
    # TODO : import directly from emoji-sequences and emoji-zwj-sequences
    # TODO : fetch data here https://unicode.org/Public/emoji/version/*.txt



import re

from .base import BasePattern
from .pattern_list import PatternTypes


class Hashtag(BasePattern):
    TYPE = PatternTypes.HASHTAG
    REGEX_FORMAT_STRING = r"#\b[^\s!@#$%^&*(),.?\":{}|<>]*\b"

    def __init__(self):
        super().__init__([re.M, re.I])



from .base import BasePattern
from .pattern_list import PatternTypes


class MacAddress(BasePattern):
    TYPE = PatternTypes.MAC_ADDRESS
    REGEX_FORMAT_STRING = r"[a-fA-F0-9]{2}(:[a-fA-F0-9]{2}){5}"



from .base import BasePattern
from .pattern_list import PatternTypes


class Uuid(BasePattern):
    TYPE = PatternTypes.UUID
    REGEX_FORMAT_STRING = (
        r"[0-9a-fA-F]{8}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-"
        r"[0-9a-fA-F]{4}\b-[0-9a-fA-F]{12}"
    )



from .base import BasePattern
from .pattern_list import PatternTypes


class HanUnification(BasePattern):
    """Detects Chinese-like characters for the 3 most popular Far East languages

    * Japanese (kanji-only)
    * Korean (hanja-only)
    * Chinese (hanzi)
    """

    TYPE = PatternTypes.HAN_UNIFICATION
    REGEX_FORMAT_STRING = (
        r"^[\u4E00-\u9FFF\u3400-\u4DBF\u20000-\u2A6DF\u2A700-"
        r"\u2B73F\u2B740-\u2B81F\u2B820-\u2CEAF\u2CEB0-\u2EBEF\u30000-\u3134F"
        r"\uF900-\uFAFF\u2E80-\u2EFF\u31C0-\u31EF\u3000-\u303F\u2FF0-\u2FFF\u3300-"
        r"\u33FF\uFE30-\uFE4F\uF900-"
        r"\uFAFF\u2F800-\u2FA1F\u3200-\u32FF\u1F200-\u1F2FF\u2F00-\u2FDF]+$"
    )



import re

from .base import BasePattern
from .pattern_list import PatternTypes


class PanFromGstin(BasePattern):
    TYPE = PatternTypes.PAN_FROM_GSTIN
    REGEX_FORMAT_STRING = (
        r"([0][1-9]|[1-2][0-9]|[3][0-5])([a-zA-Z]{5}[0-9]{4}[a-zA-Z]"
        r"{1})([1-9a-zA-Z]{1}[zZ]{1}[0-9a-zA-Z]{1})+"
    )

    def __init__(self):
        super().__init__([re.M, re.I])



import re

from .base import BasePattern
from .pattern_list import PatternTypes


class Username(BasePattern):
    """Pattern checking against any type of username with a defined length

    :param min_len: The minimal length of the username
    :type min_len: int
    :param max_len: The maximal length of the username
    :type max_len: int
    """

    TYPE = PatternTypes.USERNAME
    REGEX_FORMAT_STRING = r"\b[a-z0-9_-]{min_len,max_len}\b"

    def __init__(self, min_len: int, max_len: int):
        self.min_len = min_len
        self.max_len = max_len
        super().__init__()

    def _compile(self) -> str:
        return self.REGEX_FORMAT_STRING.replace("min_len", str(self.min_len)).replace(
            "max_len", str(self.max_len)
        )


class DiscordUsername(BasePattern):
    """Variant of the Username pattern checking against any Discord username in
    text data"""

    TYPE = PatternTypes.USERNAME
    REGEX_FORMAT_STRING = r"\b.{3,32}#[0-9]{4}\b"

    def __init__(self):
        super().__init__(flags=[re.M, re.I])



from .base import BasePattern
from .pattern_list import PatternTypes


class PortNumber(BasePattern):
    TYPE = PatternTypes.PORT_NUMBER
    REGEX_FORMAT_STRING = (
        r"\b((6553[0-5])|(655[0-2][0-9])|(65[0-4][0-9]{2})|"
        r"(6[0-4][0-9]{3})|([1-5][0-9]{4})|([0-5]{0,5})|([0-9]{1,4}))\b"
    )



from .base import BasePattern
from .pattern_list import PatternTypes


class DigitalObjectIdentifier(BasePattern):
    """Finds DOIs (Digital Object Indetifiers) in a document

    A DOI is a persistent identifier composed of a prefix/suffix,
    like 10.1000/dome.12082.
    """

    TYPE = PatternTypes.DIGITAL_OBJECT_IDENTIFIER
    REGEX_FORMAT_STRING = r"(10\.\d{4,5}\/[\S]+[^;,.\s])"



from .base import BasePattern
from .pattern_list import PatternTypes


class Ascii(BasePattern):
    """Looks for any ASCII character in a text"""

    REGEX_FORMAT_STRING = r"[ -~]"
    TYPE = PatternTypes.ASCII



from .base import BasePattern
from .pattern_list import PatternTypes


class SocialSecurityNumber(BasePattern):

    TYPE = PatternTypes.SSN
    REGEX_FORMAT_STRING = r"(?!0{3})(?!6{3})[0-8]\d{2}-(?!0{2})\d{2}-(?!0{4})\d{4}"



import re
from typing import List

from .pattern_list import PatternTypes


class BasePattern:
    """BasePattern for all pattern classes.

    Compiles into a regex pattern firsthand before processing any passed string.

    :param flags: The necessary `re` flags for the pattern compilation, defaults
    to [re.M]
    :type flags: List[re.RegexFlag], optional
    """

    REGEX_FORMAT_STRING = r""
    TYPE = PatternTypes.BASE

    @staticmethod
    def sum_regex_flags(flags: List[re.RegexFlag]) -> int:
        """Method summing regex flags for the compilation

        :param flags: list of regex flags like MULTILINE or IGNORECASE
        :type flags: List[re.RegexFlag]
        :return: An union of flags
        :rtype: int
        """
        flag_id = flags[0]
        for flag in flags[1:]:
            flag_id = flag_id | flag
        return flag_id

    def __init__(self, flags: List[re.RegexFlag] = [re.M]):
        self.flags = self.sum_regex_flags(flags)

    def _compile(self) -> str:
        """Abstract method for compile

        :return: The string to compile
        :rtype: str
        """
        return self.REGEX_FORMAT_STRING

    def compile(self) -> None:
        self.regex_pattern = re.compile(self._compile(), flags=self.flags)

    def search(self, target_string: str) -> List[str]:
        """Given some text, will look for several instances of the pattern and return a
        list of matches. Only takes the base group as the main result.

        :param target_string: The string to perform a search on.
        :type target_string: str
        :return: A list of matches or an empty list if no pattern was found
        :rtype: List[str]
        """
        if not hasattr(self, "regex_pattern"):
            self.compile()

        regex_matches = self.regex_pattern.finditer(target_string)
        return [m.group() for m in regex_matches]

    def __str__(self):
        return f"Parser for {self.TYPE}"



import re
from typing import List, Optional

from .base import BasePattern
from .pattern_list import PatternTypes


class PhoneNumber(BasePattern):
    """Pattern checking against a phone number (regular or E164)"""

    REGEX_FORMAT_STRING = r"[\+]?[(]?[0-9]{3}[)]?[-\s\.]?[0-9]{3}[-\s\.]?[0-9]{4,6}"
    REGEX_FORMAT_STRING_E164 = r"\+[1-9]\d{1,14}"
    TYPE = PatternTypes.PHONE_NUMBER

    def __init__(
        self,
        is_e164: Optional[bool] = False,
        flags: List[re.RegexFlag] = [re.M, re.IGNORECASE],
    ) -> None:
        self.is_e164 = is_e164
        super().__init__(flags=flags)

    def _compile(self) -> str:
        return (
            self.REGEX_FORMAT_STRING_E164 if self.is_e164 else self.REGEX_FORMAT_STRING
        )



from collections import namedtuple
from typing import List

from .base import BasePattern
from .pattern_list import PatternTypes

Point = namedtuple("Point", ["lat", "lon"])


class GeographicCoordinates(BasePattern):
    """Detects geographic coordinates separated by a comma within text data

    Pattern has been fitted for Earth-like coordinates"""

    TYPE = PatternTypes.GEOGRAPHIC_COORDINATES
    REGEX_FORMAT_STRING = (
        r"(\+?\b(0?[0-9]|0?[1-9][0-9]|1[0-7][0-9]|180)\b(\.\d+)?|\-"
        r"?\b(0?[0-9]|0?[1-8][0-9]|90)\b(\.\d+)?),\s*(\+?\b(0?[0-9]"
        r"|0?[1-9][0-9]|[12][0-9][0-9]|3[0-5][0-9]|360)\b(\.\d+)?|"
        r"\-?\b(0?[0-9]|0?[1-9][0-9]|1[0-7][0-9]|180)\b(\.\d+)?)"
    )

    def to_coords(self, target_string: str) -> List[Point]:
        """Returns the parsed coordinates as an object with the latitude
        and longitude as floats

        :param target_string: the string to parse
        :type target_string: str
        :return: A list of coordinate objects Point(lat, lon)
        :rtype: List[Point]
        """
        matches = self.search(target_string)
        points = []

        for match in matches:
            lat, lon = match.split(",")
            points.append(Point(float(lat.strip(" +")), float(lon.strip(" +"))))

        return points



from enum import Enum


class PatternTypes(str, Enum):
    BASE = "base"
    USERNAME = "username"
    PASSWORD = "password"
    EMAIL = "email"
    PHONE_NUMBER = "phone number"
    IP_ADDRESS = "IP adress"
    DATE = "date"
    ASCII = "ascii"
    URL = "url"
    SSN = "ssn"
    SEMANTIC_VERSIONING = "semantic versioning"
    UUID = "UUID"
    JIRA_ISSUE_TICKET = "jira issue ticket"
    GEOGRAPHIC_COORDINATES = "geographic coordinates"
    MAC_ADDRESS = "mac address"
    URL_SLUG = "url slug"
    C_IDENTIFIER = "C identifier"
    HYPHEN_WORD_BREAK = "hyphen word break"
    BITCOIN_ADDRESS = "bitcoin address"
    EMOJI = "emoji"
    CREDIT_CARD_NUMBER = "credit card number"
    PAN_FROM_GSTIN = "PAN from GSTIN"
    HAN_UNIFICATION = "han unification"
    PORT_NUMBER = "port number"
    DIGITAL_OBJECT_IDENTIFIER = "digital object identifier"
    HASHTAG = "hashtag"



from .base import BasePattern
from .pattern_list import PatternTypes


class Email(BasePattern):
    """
    Looks for any type of email in the text data.

    Note: the pattern used here is quite complex. You can look up here
    at https://ihateregex.io/expr/email-2/
    """

    REGEX_FORMAT_STRING = (
        r"""(([^<>()\[\]\\.,;:\s@"]+(\.[^<>()\[\]\\.,;:\s@"]+)*)|(".+")"""
        r""")@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}])|(([a-zA-Z\-0-9]"""
        r"""+\.)+[a-zA-Z]{2,}))"""
    )
    TYPE = PatternTypes.EMAIL



from .base import BasePattern
from .pattern_list import PatternTypes


class SemanticVersioning(BasePattern):
    TYPE = PatternTypes.SEMANTIC_VERSIONING
    REGEX_FORMAT_STRING = (
        r"(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]"
        r"\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*"
        r"[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\."
        r"[0-9a-zA-Z-]+)*))?"
    )



from .base import BasePattern
from .pattern_list import PatternTypes


class Password(BasePattern):
    TYPE = PatternTypes.PASSWORD
    REGEX_FORMAT_STRING = (
        r"(?=.*?[A-Z])(?=.*?[a-z])(?=.*?[0-9])(?=.*?[#?!@$ %^&*-]).{min_len,}"
    )

    def __init__(self, min_len: int):
        super().__init__()
        self.min_len = min_len

    def _compile(self) -> str:
        return self.REGEX_FORMAT_STRING.replace("min_len", str(self.min_len))



from .ascii import Ascii
from .bitcoin_address import BitcoinAddress
from .c_identifier import CLanguageIdentifier
from .coords import GeographicCoordinates
from .credit_card import CreditCardNumber
from .date import Date
from .doi import DigitalObjectIdentifier
from .email import Email
from .emoji import Emoji
from .han_unification import HanUnification
from .hashtag import Hashtag
from .hyphen_word_break import HyphenWordBreak
from .ip import IpAddress
from .jira import JiraTicket
from .mac_address import MacAddress
from .pan_gstin import PanFromGstin
from .password import Password
from .phone_number import PhoneNumber
from .port_number import PortNumber
from .semantic_versioning import SemanticVersioning
from .ssn import SocialSecurityNumber
from .url import Url
from .url_slug import UrlSlug
from .username import DiscordUsername, Username
from .uuid import Uuid
